#include "cnf_stmt.h"

#include "SATInstance.h"

#include <algorithm>
#include <string>
#include <vector>
#include <hash_map>
#include <strstream.h>

#include <ctype.h>
#include <stdlib.h>

// a <= b if a's number is < b's or a's # == b's # and a is
// positive or both a and b are negative.
struct Compare_VAR_OCC {
  bool operator() (const VAR_OCC & a, const VAR_OCC & b) {
    return (a.val < b.val || 
		 (a.val == b.val &&
		  (a.pos || !b.pos)));
  } 
};



void tokenize(string & str)
{
  unsigned int pos;
  while ((pos = str.find(CONNECTOR)) <= str.length()) {
    str.replace(pos, strlen(CONNECTOR), " ");
  }
}

bool is_constraint(string token)
{
  const string CPP_SET_STR = SET_STR;
  const string CPP_TEST_STR = TEST_STR;

  if (token[0] == '!')
    token = token.substr(1);
  
  return 
    (token.compare(CPP_SET_STR, 0, CPP_SET_STR.length()) == 0 ||
	token.compare(CPP_TEST_STR, 0, CPP_TEST_STR.length()) == 0);
}

int find_max_var(const string & cons)
{
  unsigned int pos, space_pos;
  int max = 0, tmp;

  pos = cons.find("r");

  while (pos < cons.length()) {
    pos++;
    space_pos = pos;
    while (space_pos < cons.length() && isdigit(cons[space_pos]))
	 space_pos++;

    tmp = atoi(cons.substr(pos, space_pos).c_str());
    if (tmp > max)
	 max = tmp;

    pos = cons.find("r", pos);
  }

  return max;
}

void CNF_STMT::prettyPrint(ostream & out)
{
  vector<CNF_CLAUSE>::iterator i;
  CNF_CLAUSE::iterator j;

  for (i = my_clauses.begin(); i != my_clauses.end(); i++) {
    if ((*i).empty())
	 continue;
    for (j = (*i).begin(); j != (*i).end(); j++) {
	 out << ((*j).pos ? "" : "!") << my_map.second[(*j).val] << " ";
    }
    out << endl;
  }
}

void CNF_STMT::print(ostream & outCNF, ostream & outMap)
{
  vector<CNF_CLAUSE>::iterator i;
  CNF_CLAUSE::iterator j;
  hash_map<int, string>::iterator ci;
  hash_map<string, int>::iterator mi;

  // Output the CNF file
  outCNF << "c LCNF file produced by LiPSyNC from " << problemName << "." << endl;
  outCNF << "p lcnf " << my_map.second.size() << " " << my_clauses.size() << " "
	 << my_vmap.second.size() << endl;
  for (i = my_clauses.begin(); i != my_clauses.end(); i++) {
    for (j = (*i).begin(); j != (*i).end(); j++) {
	 outCNF << ((*j).pos ? "" : "-") << (*j).val + 1 << " ";
    }
    outCNF << "0" << endl;
  }
  
  for (ci = my_constraints.begin(); ci != my_constraints.end(); ci++) {
    outCNF << (*ci).first + 1 << " [ " << (*ci).second << " ]" << endl;
  }
  outCNF << "0" << endl;

  // Output the map file
  // Form is:
  // file => <vars> 0 <rvars> 0
  // vars => <var>*
  // var => # name
  // rvars => <var>*
  for (mi = my_map.first.begin(); mi != my_map.first.end(); mi++) {
    if ((*mi).second >= 0)
	 outMap << (*mi).second + 1 << " " << (*mi).first << endl;
    else
	 outMap << (*mi).second << " " << (*mi).first << endl;
  }
  outMap << 0 << endl;  
  for (mi = my_vmap.first.begin(); mi != my_vmap.first.end(); mi++) {
    if ((*mi).second >= 0)
	 outMap << (*mi).second + 1 << " " << (*mi).first << endl;
    else
	 outMap << (*mi).second << " " << (*mi).first << endl;
  }
  outMap << 0 << endl;  
}

void CNF_STMT::set(fact_list stmt)
{
  while (stmt) {
    add(stmt->item);
    stmt = stmt->next;
  }
}

// add an untokenized clause
void CNF_STMT::add(fact_list clause)
{
  string tok;
  int i_tok;
  CNF_CLAUSE c;
  bool trivial;
  MAPPING_VALUE mvCurr;

  if (detail_level >= DETAIL_ALL_DEBUG) {
    print_clauselist(clause);
    cout << endl;
  }

  while (clause) {
    assert(get_adl_token(clause) == OR_CONST);
    
    clause = clause->next;
    
    c.clear();
    trivial = false;

    while (get_adl_token(clause) != RBRACK_CONST) {
	 assert(clause->item->item);
	 tok = clause->item->item;

	 for (token_list t = clause->item->next; t; t = t->next) {
	   tok += ((string)CONNECTOR) + t->item;
	 }

	 add_token(tok);
	 i_tok = find_token(tok);
	 mvCurr = tok[0] == '!' ? MV_FALSE : MV_TRUE;

	 if (i_tok < 0) {
	   if (i_tok == mvCurr || i_tok == MV_CONTRADICTION) {
		while (get_adl_token(clause) != RBRACK_CONST)
		  clause = clause->next;
		trivial = true;
		break;
	   }
	 }
	 else {
	   c.push_back(VAR_OCC(tok[0] != '!', i_tok));
	 }

	 clause = clause->next;
    }

    if (!trivial)
	 add(c);

    clause = clause->next;
  }
}

void CNF_STMT::readSolution(istream & inSol, istream & inMap,
					   hash_map<string, bool> & soln, 
					   hash_map<string, double> & vsoln)
{
  // Read the map into the hash_tables
  int var;
  string name;
  char check;

  while (inMap >> var && var != 0) {
    inMap >> name;
    if (var > 0) {
	 var--;
	 if (var >= (int)my_map.second.size())
	   my_map.second.resize(var + 1);
	 my_map.second[var] = name;
	 my_map.first[name] = var;
    }
    else {
	 soln[name] = (var == MV_TRUE) ? true : false;
    }
  }

  while (inMap >> var && var != 0) {
    inMap >> name;
    if (var > 0) {
	 var--;
	 if (var >= (int)my_vmap.second.size())
	   my_vmap.second.resize(var + 1);
	 my_vmap.second[var] = name;
	 my_vmap.first[name] = var;    
    }
    else {
	 vsoln[name] = 0;
    }
  }

  // Read the solution
  while (inSol >> check && check != '0') {
    bool bPos = true;
    if (check == '-')
	 bPos = false;
    else
	 inSol.putback(check);

    inSol >> var;
    var--;

    soln[my_map.second[var]] = bPos;
  }

  while (inSol >> var && var != 0) {
    double val;
    inSol >> val;
    var--;

    vsoln[my_vmap.second[var]] = val;
  }
}

void CNF_STMT::add(token_list clause)
{
  string tok;
  CNF_CLAUSE c;
  int i_tok;

  while (clause) {
    tok = clause->item;
    add_token(tok);
    i_tok = find_token(tok);

    c.push_back(VAR_OCC(tok[0] != '!', i_tok));

    clause = clause->next;
  }

  add(c);
}

void CNF_STMT::add_token(string & str)
{
  bool pos = true;

  if (str[0] == '!') {
    str = str.substr(1);
    pos = false;
  }

  bool is_cons = is_constraint(str);

  if (is_cons) {
    str = str.substr(str.find(CONNECTOR));
    str = CONSTRAINT_STR + str;
  }

  if (my_map.first.find(str) == my_map.first.end()) {
    my_map.first[str] = my_map.second.size();
    my_map.second.push_back(str);
    my_vars.push_back(list<pair<int,bool> >());

    if (is_cons) {
	 add_constraint(str);
    }
  }

  if (!pos)
    str = "!" + str;
}

void CNF_STMT::add_constraint(string str)
{
  int i = find_token(str);
  string var;
  char buf[MAX_LENGTH];

  tokenize(str);

  istrstream cons_in(str.c_str());
  ostrstream cons_out(buf, MAX_LENGTH);

  // Ditch the SET/TEST
  cons_in >> var;

  while (cons_in >> var) {
    // Count and file the rvars
    if (var[0] == '?') {
	 add_var(var);

	 cons_out << "r" << (find_var(var) + 1) << " ";
    }
    else {
	 cons_out << var << " ";
    }
  } 
  cons_out << '\0';

  my_constraints[i] = buf;
}

void CNF_STMT::add_var(string var)
{
  assert(var[0] == '?');

  var = var.substr(1);

  if (my_vmap.first.find(var) == my_vmap.first.end()) {
    my_vmap.first[var] = my_vmap.second.size();
    my_vmap.second.push_back(var);
  }
}

int CNF_STMT::find_var(string var)
{
  assert(var[0] == '?');
  
  var = var.substr(1);

  assert(my_vmap.first.find(var) != my_vmap.first.end());

  return my_vmap.first[var];
}

string CNF_STMT::find_var(int var)
{
  assert(var >= 0 && (unsigned int)var < my_vmap.second.size());

  return my_vmap.second[var];
}

int CNF_STMT::find_token(string str)
{
  if (str[0] == '!')
    str = str.substr(1);

  assert(my_map.first.find(str) != my_map.first.end());

  return (my_map.first[str]);
}

string CNF_STMT::find_token(int i)
{
  assert(i >= 0 && (unsigned int)i < my_map.second.size());

  return my_map.second[i];
}

void CNF_STMT::add(CNF_CLAUSE c)
{
  Compare_VAR_OCC comp;
  sort<CNF_CLAUSE::iterator, Compare_VAR_OCC>(c.begin(), c.end(), comp);

  if (c.size() == 0) {
    if (my_contradiction < 0)
	 my_contradiction = 0;
    return;
  }

  // weed out silly stuff like repeated lits and trivially satisfiable
  // clauses
  streamline_clause(c);

  // check for trivial clauses
  if (c.size() == 0)
    return;

  // check for unit clauses
  if (c.size() == 1) {
    unit_vars[c[0].val] = c[0].pos;
    unit_propagate();
    return;
  }

  // enter this clause into its variables' clause lists
  for (unsigned int i = 0; i < c.size(); i++) {
    my_vars[c[i].val].push_back(pair<int,bool>(my_clauses.size(), c[i].pos));
  }

  // add the clause to the clause list
  my_clauses.push_back(c);
}

void CNF_STMT::streamline_clause(CNF_CLAUSE & c)
{
  int last_var;
  int j;

  last_var = -1;
  for (j = 0; j < (signed int)c.size(); j++) {
    // any var repeated (either positively or negatively) is a problem
    if (c[j].val == last_var) {
	 
	 // repeated literal: destroy one copy
	 if (c[j-1].pos == c[j].pos) {
	   c.erase(c.begin() + j);
	   j--;
	 }
	 // negated literals: destroy this clause (trivially satisfiable)
	 else {
	   c.clear();
	   break;
	 }
    }
    else {
	 last_var = c[j].val;
    }
  }
}

void CNF_STMT::unit_propagate()
{
  int var;
  bool var_pos;

  while (!unit_vars.empty()) {
    var = (*unit_vars.begin()).first;
    var_pos = (*unit_vars.begin()).second;

    set(VAR_OCC(var_pos, var));
    unit_vars.erase(var);
  }
}

void CNF_STMT::set(VAR_OCC lit)
{
  int var = lit.val;
  bool var_pos = lit.pos;
  int clause;
  bool clause_pos;
  string var_str = my_map.second[var];

  my_map.first[var_str] = lit.pos ? MV_TRUE : MV_FALSE;


  for (list<pair<int, bool> >::iterator li = my_vars[var].begin();
	  li != my_vars[var].end();
	  li++) {
    clause = (*li).first;
    clause_pos = (*li).second;

    // check if the clause has already been satisfied
    if (my_clauses[clause].empty())
	 continue;

    // satisfy clause
    if (clause_pos == var_pos) {
	 my_clauses[clause].clear();
    }
    // filter clause
    else {
	 CNF_CLAUSE::iterator ci;
	 ci = my_clauses[clause].begin();
	 while (ci != my_clauses[clause].end()) {
	   if ((*ci).val == var)
		break;
	   else if ((*ci).val > var) {
		ci = my_clauses[clause].end();
		break;
	   }
	   ci++;
	 }
	 if (ci != my_clauses[clause].end()) {
	   my_clauses[clause].erase(ci);

	   // check for unit clause
	   if (my_clauses[clause].size() == 1) {
		unit_vars[my_clauses[clause][0].val] = my_clauses[clause][0].pos;
	   }
		
	   // check for contradiction:
	   if (my_clauses[clause].size() == 0) {
		my_contradiction = var;
		my_map.first[var_str] = MV_CONTRADICTION;
	   }
	 }
    }
  }
}

void CNF_STMT::eliminate_pure_literals()
{
  VAR_OCC v;

  for (v = find_pure_lit(); v.val != -1; v = find_pure_lit()) {
    set(v);
  }
}

VAR_OCC CNF_STMT::find_pure_lit()
{
  static hash_map<int, pair<int, int> > count;
  VAR_OCC var(true, -1);
  int erase;

  if (count.empty()) {
    for (unsigned int i = 0; i < my_clauses.size(); i++) {
	 for (unsigned int j = 0; j < my_clauses[i].size(); j++) {
	   int val = my_clauses[i][j].val;
	   if (count.find(val) == count.end()) {
		count[val].first = 0;
		count[val].second = 0;
	   }
	   
	   if (my_clauses[i][j].pos) {
		count[val].first++;
	   }
	   else {
		count[val].second++;
	   }
	 }
    }
    
    erase = -1;
    for (hash_map<int, pair<int, int> >::iterator mi = count.begin();
	    mi != count.end(); mi++) {
	 if (erase >= 0) {
	   count.erase(erase);
	   erase = -1;
	 }
	 int val = (*mi).first, npos = (*mi).second.first, 
	   nneg = (*mi).second.second;
	 if (npos == 0) {
	   var.pos = false;
	   var.val = val;
	 }
	 else if (nneg == 0 && my_constraints.find(val) == my_constraints.end()) {
	   var.pos = true;
	   var.val = val;
	 }
	 else {
	   erase = val;
	 }
    }

    if (erase >= 0)
	 count.erase(erase);

    return var;
  }    
  else {
    hash_map<int, pair<int,int> >::iterator mi = count.begin();
    var.val = (*mi).first;
    var.pos = ((*mi).second.first == 0) ? false : true;
    count.erase(mi);
    
    return var;
  }
}

void CNF_STMT::renumber()
{
  vector<CNF_CLAUSE>::iterator cnf_i;
  CNF_CLAUSE::iterator clause_i;

  hash_map<int, string>::iterator ci;
  hash_map<string, int>::iterator mi;
  hash_map<int, string> new_constraints;
  vector<string> new_map;
  vector<int> renumber_map;
  vector<int> renumber_vmap;
  list<int> delete_list;
  int i, j;

  if (detail_level >= DETAIL_ALL_DEBUG)
    cerr << "Beginning renumbering; thinning clauses" << endl;

  // first, clear out satisfied (empty) clauses:
  for (i = (signed int)my_clauses.size() - 1; i >= 0; i--) {
    if (my_clauses[i].empty()) {
	 j = i - 1;
	 while (j >= 0 && my_clauses[j].empty())
	   j--;
	 my_clauses.erase(my_clauses.begin() + j + 1, my_clauses.begin() + i + 1);
	 i = j;
    }
  }

  if (detail_level >= DETAIL_ALL_DEBUG)
    cerr << "Thinned clauses; identifying stray variables" << endl;

  // then, set variables which appear nowhere to don't care.
  for (i = 0; i < (signed int)my_map.first.size(); i++) {
    if (my_map.first[my_map.second[i]] == i)
	 my_map.first[my_map.second[i]] = MV_DONT_CARE;
  }

  for (cnf_i = my_clauses.begin(); cnf_i != my_clauses.end(); cnf_i++) {
    for (clause_i = (*cnf_i).begin(); clause_i != (*cnf_i).end(); clause_i++) {
	 if (my_map.first[my_map.second[(*clause_i).val]] == MV_DONT_CARE) {
	   my_map.first[my_map.second[(*clause_i).val]] = (*clause_i).val;
	 }
    }
  }

  if (detail_level >= DETAIL_ALL_DEBUG)
    cerr << "Identified stray variables; handling contradictions" << endl;

  // next, add contradiction clauses if necessary:
  if (my_contradiction >= 0) {
    CNF_CLAUSE p, n;
    p.push_back(VAR_OCC(true, my_contradiction));
    n.push_back(VAR_OCC(false, my_contradiction));

    my_clauses.push_back(p);
    my_clauses.push_back(n);

    // make sure this variable isn't thrown away (as already assigned)
    my_map.first[my_map.second[my_contradiction]] = my_contradiction;
  }

  if (detail_level >= DETAIL_ALL_DEBUG)
    cerr << "Handled contradictions; restoring positive constraints" << endl;

  // next, add unit positive constrained preds back in (to
  // make sure the solver registers the constraints).
  for (ci = my_constraints.begin(); ci != my_constraints.end(); ci++) {
    MAPPING_VALUE mv = (MAPPING_VALUE)my_map.first[my_map.second[(*ci).first]];
    if (mv == MV_TRUE) {
	 CNF_CLAUSE c;

	 // add a unit clause
	 c.push_back(VAR_OCC(true, (*ci).first));
	 my_clauses.push_back(c);

	 // make sure this variable isn't thrown away (as already assigned)
	 my_map.first[my_map.second[(*ci).first]] = (*ci).first;
    }
  }

  if (detail_level >= DETAIL_ALL_DEBUG)
    cerr << "Restored positive constraints; renumbering pred/cons" << endl;

  renumber_map.resize(my_map.second.size());
  renumber_vmap.resize(my_vmap.second.size());

  // now, find the predicate renumbering (simultaneously, renumber 
  // constraints)
  for (i = 0; i < (signed int)my_map.second.size(); i++) {
    if (my_map.first[my_map.second[i]] == i) {
	 if (my_constraints.find(i) != my_constraints.end()) {
	   new_constraints[new_map.size()] = my_constraints[i];
	 }

	 renumber_map[i] = new_map.size();

	 my_map.first[my_map.second[i]] = new_map.size();
	 new_map.push_back(my_map.second[i]);
    }
    else {
	 assert(my_map.first[my_map.second[i]] < 0);
    }
  }

  if (detail_level >= DETAIL_ALL_DEBUG)
    cerr << "Renumbered pred/cons; applying renumbering" << endl;

  // then, set the renumbering as the real numbering
  
  // both in the maps/constraints
  my_map.second = new_map;
  my_constraints = new_constraints;

  // and in the clauses themselves
  for (cnf_i = my_clauses.begin();
	  cnf_i != my_clauses.end();
	  cnf_i++) {
    for (clause_i = (*cnf_i).begin();
	    clause_i != (*cnf_i).end();
	    clause_i++) {
	 (*clause_i).val = renumber_map[(*clause_i).val];
    }
  }

  new_map.clear();

  // next, renumber the vars (including setting unneeded vars
  // to MV_DONT_CARE)
  for (mi = my_vmap.first.begin(); mi != my_vmap.first.end(); mi++)
    (*mi).second = MV_DONT_CARE;

  for (ci = my_constraints.begin(); ci != my_constraints.end(); ci++) {
    char buf[MAX_LENGTH];
    char c;
    ostrstream cons_out(buf, MAX_LENGTH);
    istrstream cons_in((*ci).second.c_str());
    int i_tok;

    while (cons_in.get(c)) {
	 cons_out << c;
	 if (c == 'r') {
	   cons_in >> i_tok;
	   i_tok--;
	   
	   if (my_vmap.first[my_vmap.second[i_tok]] == MV_DONT_CARE) {
		renumber_vmap[i_tok] = new_map.size();
		my_vmap.first[my_vmap.second[i_tok]] = new_map.size();
		new_map.push_back(my_vmap.second[i_tok]);
	   }

	   i_tok = renumber_vmap[i_tok];

	   cons_out << i_tok + 1;
	 }
    }
    cons_out << '\0';

    (*ci).second = buf;
  }

  my_vmap.second = new_map;

  if (detail_level >= DETAIL_ALL_DEBUG)
    cerr << "Done renumbering" << endl;
}

bool CNF_STMT::subsumes(CNF_CLAUSE small, CNF_CLAUSE big)
{
  int i = 0, j = 0;

  while (i < (signed int)small.size() && j < (signed int)small.size()) {
    if (small[i].val == big[j].val && small[i].pos == big[i].pos) {
	 i++;
	 j++;
    }
    else if (small[i].val <= big[j].val) {
	 // no way to get back to where big[j].val < small[i].val without
	 // skipping a small value
	 return false;
    }
    else {
	 j++;
    }
  }
  // make sure all small values were accounted for.
  return i == (signed int)small.size();
}

void CNF_STMT::subsume()
{
  int i, j;
  for (i = 0; i < (signed int)my_clauses.size(); i++) {
    if (my_clauses[i].empty())
	 continue;
    for (j = i + 1; j < (signed int)my_clauses.size(); j++) {
	 if (my_clauses[j].empty())
	   continue;
	 if (my_clauses[i].size() <= my_clauses[j].size()) {
	   if (subsumes(my_clauses[i], my_clauses[j])) {
		my_clauses[j].clear();
	   }
	 }
	 else {
	   if (subsumes(my_clauses[j], my_clauses[i])) {
		my_clauses[i].clear();
		break;
	   }
	 }
    }
  }
}

int CNF_STMT::size()
{
  int size = 0;
  for (unsigned int i = 0; i < my_clauses.size(); i++)
    if (!my_clauses[i].empty())
	 size++;

  return size;
}

void CNF_STMT::compact()
{
  if (detail_level >= DETAIL_STATS) {
    cerr << "Initial statement stats:" << endl;
    cerr << "\tNumber of clauses: " << size() << endl;
    cerr << "\tNumber of variables: " << my_map.second.size() << endl;
    cerr << "\tNumber of constraints: " << my_constraints.size() << endl;
    cerr << "\tNumber of constraint variables: " << my_vmap.second.size() 
	    << endl;
  }

  if (detail_level >= DETAIL_MINIMAL_DEBUG) {
    if (compact_level >= COMPACT_UNIT) {
	 cerr << "\t\t# of clauses: " << size() << endl;
	 cerr << "\tUnit propagating" << endl;
	 unit_propagate();
    }
    
    if (compact_level >= COMPACT_SUBSUME) {
	 cerr << "\t\t# of clauses: " << size() << endl;
	 cerr << "\tSubsuming" << endl;
	 subsume();
    }
    
    if (compact_level >= COMPACT_PURE) {
	 cerr << "\t\t# of clauses: " << size() << endl;
	 cerr << "\tEliminating pure literals" << endl;
	 eliminate_pure_literals();
    }

    cerr << "\t\t# of clauses: " << size() << endl;
    cerr << "\tRenumbering" << endl;
    renumber();
  }
  else {
    if (compact_level >= COMPACT_UNIT)
	 unit_propagate();
    if (compact_level >= COMPACT_SUBSUME)
	 subsume();
    if (compact_level >= COMPACT_PURE)
	 eliminate_pure_literals();
    renumber();
  }

  if (detail_level >= DETAIL_STATS) {
    cerr << "Final statement stats:" << endl;
    cerr << "\tNumber of clauses: " << my_clauses.size() << endl;
    cerr << "\tNumber of variables: " << my_map.second.size() << endl;
    cerr << "\tNumber of constraints: " << my_constraints.size() << endl;
    cerr << "\tNumber of constraint variables: " << my_vmap.second.size() 
	    << endl;
  }
}

bool CNF_STMT::solve(hash_map<string, bool> & soln, 
				 hash_map<string, double> & vsoln,
				 bool & timedOut)
{
  boolean solved = false; 
  hash_map<string, int>::iterator mi;
  vector< pair<int, string> > constraints;
  int max_var = 0;
  timedOut = false;
  SATSolver my_solver;

  if (my_contradiction >= 0) {
    return false;
  }

  for (hash_map<int, string>::iterator i = my_constraints.begin();
	  i != my_constraints.end();
	  i++) {
    int this_max = find_max_var((*i).second);
    if (this_max > max_var)
	 max_var = this_max;
    constraints.push_back(*i);
  }

  assert(max_var == my_vmap.second.size());

  SATInstance::xSATInstance.vDestroy();
  if (!SATInstance::xSATInstance.bInit(my_clauses, my_map.second.size(),
							    constraints, my_vmap.second.size()))
    return false;

  solved = my_solver.bFindSolution(timedOut);

  if (timedOut) {
    cerr << "TIME LIMIT EXPIRED" << endl;
    return false;
  }

  if (!solved) {
    return false;
  }

  for (mi = my_map.first.begin(); mi != my_map.first.end(); mi++) {
    if ((*mi).second < 0) {
	 soln[(*mi).first] = (*mi).second == MV_TRUE;
    }
    else {
	 soln[(*mi).first] = my_solver.bGetValue((*mi).second);
    }
  }

  for (mi = my_vmap.first.begin(); mi != my_vmap.first.end(); mi++) {
    if ((*mi).second < 0) {
	 vsoln[(*mi).first] = 0;
    }
    else {
	 vsoln[(*mi).first] = my_solver.dGetValue((*mi).second);
    }
  }  

  return true;
}

