/* (C) Copyright 1997 Albert Ludwigs University Freiburg
 *     Institute of Computer Science
 *
 * All rights reserved. Use of this software is permitted for 
 * non-commercial research purposes, and it may be copied only 
 * for that use.  All copies must include this copyright message.
 * This software is made available AS IS, and neither the authors
 * nor the  Albert Ludwigs University Freiburg make any warranty
 * about the software or its performance. 
 */


#include <time.h>
#include <math.h>
#include <fstream.h>

#include "ipp.h" /* defines, data structures, fn prototypes, 
				global variables */
#include "cnf_stmt.h" /* for the CNF_STMT type */
#include "Args.h" /* for argument processing */
#include "Random.h" /* to initialize random numbers */
#include <hash_map>
#include <string>
#include "stats.h"
#include "SATInstance.h"
#include "SATSolver.h"

/*
 * global variables
 */


/* technical variables */
char *act_filename;
int lineno = 1;
hash_map<string, Stats> stats;

/* globally important variables */
int detail_level = DETAIL_MINIMAL;
int compact_level = COMPACT_ALL;
int cache_policy = NO_CACHE;
int cache_depth = 0;
bool bParallel = true;
bool bInfluences = true;
bool bMutexInfl = true;

/* domain constants */
char *gdomain_name;
int gproblem_length = 1;/* actual problem length (decided after args set) */
int gproblem_length_parallel = -1;/* problem length if parallel */
int gproblem_length_serial = -1;/* problem length if serial */
op_list loaded_ops = NULL;/* loaded, uninstantiated operators */
operator_list operators = NULL;/* list of instantiated operators */
token_list initial_facts = NULL;/* the initial facts */
fact_list raw_goal_facts = NULL; /* the goal facts as a clause list */
fact_list orig_constant_list = NULL;/* to store all typed objects */
fact_list orig_initial_facts = NULL;/* stores initials as fact_list */
fact_list domain_vars_list = NULL;/* list of all global cvars in the domain */
fact_list predicates_list = NULL;/* list of all predicates w/types in domain */
fact_list global_object_fl = NULL;/* fast finding of types and their objs */
type_tree_list global_type_tree_list = NULL;/* type hierarchy (PDDL) */
CNF_STMT solver_stmt;

char problemName[MAX_LENGTH]; /* the problem name */

bool bOutputOnly = false;
bool bLoadOnly = false;
bool bSolveOnly = false;
const string cnfExt = ".cnf";
const string mapExt = ".map";
const string outExt = ".out";
string cnfFile = "";
string mapFile = "";
string outFile = "";
int randomSeed;
bool bTimeLimit;
int iTimeLimit;
bool bRandomRestart;
int iInitLimit;
double dRRMultiplier = 2;
double dRRIncrement = 0;
Args argParser;

char ops_file_name[MAX_LENGTH] = "";
char fct_file_name[MAX_LENGTH] = "";
char path[MAX_LENGTH]  = "";/* common path for both files */
char ops_file[MAX_LENGTH]  = "";/* resulting name for ops file */
char fct_file[MAX_LENGTH]  = "";/* same for fct file */
BOOLEAN found_plan = FALSE;/* stores return value of search_plan */

void argSetup();
void argGather();
void argPrint();

void parse()
{
  /* add path info, complete file names will be stored in
   * ops_file and fct_file
   */
  strcat( ops_file, path );
  strcat( fct_file, path );
  strcat( ops_file, ops_file_name );
  strcat( fct_file, fct_file_name );

  get_fct_file_name( fct_file );

  if (detail_level >= DETAIL_MINIMAL_DEBUG) {
    cout << "Loading files." << endl;
  }

  load_ops_file( ops_file ); /* it is important for the pddl language
						  to define the domain before reading
						  the problem */

  load_fct_file( fct_file );

  // Now that the length hints are in, choose the appropriate hint to use.
  if (((IntArg*)argParser.getArg("steps"))->val < 0) {
    if (bParallel) {
	 if (gproblem_length_parallel >= 0) {
	   gproblem_length = gproblem_length_parallel;
	 }
	 else if (detail_level >= DETAIL_MINIMAL) {
	   cout << "Using default problem length since no parallel length" 
		   << " specified" << endl;
	 }
    }
    else {
	 if (gproblem_length_serial >= 0) {
	   gproblem_length = gproblem_length_serial;
	 }
	 else if (detail_level >= DETAIL_MINIMAL) {
	   cout << "Using default problem length since no serial length specified"
		   << endl;
	 }
    }
  }
  else {
    gproblem_length = ((IntArg*)argParser.getArg("steps"))->val;
  }
  
  if (detail_level >= DETAIL_MINIMAL_DEBUG)
    cout << "Using problem length: " << gproblem_length << endl;
}

void preprocess()
{
  if (detail_level >= DETAIL_MINIMAL_DEBUG)
    cout << "Preprocessing." << endl;

  build_orig_constant_list();

  preprocess_pl1_facts(); /* find normal forms for pl1 expression
					    in action hooks and goals */

  /* calculate all possible instantiations 
   * and store them in global list operators
   */
  instantiate_list( loaded_ops, orig_constant_list );  

  eliminate_timeless(&raw_goal_facts);

  preprocess_domain_vars(); /* handle initial values for domain vars */

  transform_initials(); /* now closed world assumption is accounted for */
}

void output(SATSolver & solver, ostream & out)
{
  if (found_plan) {
    solver.vOutput(out);
  }
  else {
    out << "NO SOLUTION" << endl;
  }
}

bool solve(SATSolver & solver, bool & bTimedOut)
{
  bool bSolved;

  bTimedOut = false;
  SATInstance::xSATInstance.vDestroy();
  if (!SATInstance::xSATInstance.bReadDimacs((char*)cnfFile.c_str()))
    return false;

  bSolved = solver.bFindSolution(bTimedOut);

  if (bTimedOut) {
    cerr << "TIME LIMIT EXPIRED" << endl;
    return false;
  }

  return bSolved;
}

/********************************************************************************/
/* Main function                                                                */
/********************************************************************************/

int main( int argc, char *argv[] )
{
  hash_map<string, bool> solution;
  hash_map<string, double> rsolution;
  
  // Set up the argument parser
  argSetup();

  // Parse the arguments:
  if (!argParser.parseArgs(argc, argv))
    exit(1);

  // Gather the results:
  argGather();

  if (detail_level >= DETAIL_STATS)
    argPrint();

  Random::vInitRandom(randomSeed);

  log_start(TOTAL_TIME, DETAIL_MINIMAL);

  // Silliness
  if (((IntArg*)argParser.getArg("detail"))->val == DETAIL_FLIG) {
    for (int i = 0; i < 1002; i++) {
	 if (i >= 234 && i < 236)
	   cerr << "BONUS ";
	 cerr << "FLIG" << endl;
    }
    exit(1);
  }

  if (!bLoadOnly && !bSolveOnly) {
    log_start("Parsing", DETAIL_STATS);
    parse();
    log_end("Parsing");
    
    log_start("Preprocessing", DETAIL_STATS);
    preprocess();
    log_end("Preprocessing");
    
    log_start("Build CNF", DETAIL_STATS);
    build_cnf_stmt(); /* builds solver_stmt */
    log_end("Build CNF");
    
    log_start("Compact", DETAIL_STATS);
    
    /* Free up all memory still held by ipp */
    my_free_all();
    
    if (detail_level >= DETAIL_ALL_DEBUG) {
	 cout << "CNF_STMT before compact: " << endl;
	 solver_stmt.prettyPrint();
    }
    
    if (detail_level >= DETAIL_MINIMAL_DEBUG)
	 cout << "Compacting." << endl;
    
    solver_stmt.compact();
    
    if (detail_level >= DETAIL_ALL_DEBUG) {
	 cout << "CNF_STMT after compact: " << endl;
	 solver_stmt.prettyPrint();
    }
    
    log_end("Compact");
  }

  /* Check for printout request */
  if (bOutputOnly) {
    ofstream outCNF(cnfFile.c_str());
    ofstream outMap(mapFile.c_str());
    solver_stmt.print(outCNF, outMap);
    log_end(TOTAL_TIME);
    wrap_up();
    exit(0);
  }

  SATSolver * solver = 0;
  
  if (!bLoadOnly && !bOutputOnly) {
    /* Solve, with random restarts if necessary. */
    bool bRestarted = false;
    bool bTimedOut;

    do {
	 if (solver && bRestarted)
	   delete solver;
	 
	 /* each rep of random restart, I double the initial time limit. */
	 if (iInitLimit + logged_time("Solving") > iTimeLimit)
	   SATSolver::iMaxTime = iTimeLimit - logged_time("Solving");
	 else
	   SATSolver::iMaxTime = iInitLimit;
	 iInitLimit = (int)ceil(dRRMultiplier * iInitLimit + dRRIncrement);
	 
	 log_start("Solving", DETAIL_STATS);
	 
	 if (detail_level >= DETAIL_MINIMAL_DEBUG)
	   cout << "Solving." << endl;

	 if (bSolveOnly) {
	   solver = new SATSolver();
	   found_plan = solve(*solver, bTimedOut);
	 }	   
	 else
	   found_plan = solver_stmt.solve(solution, rsolution, bTimedOut); 	   
	 
	 log_end("Solving");
	 
	 if (bRestarted)
	   stats["Restarts"].log();
	 else
	   bRestarted = true;
    } while (bRandomRestart && bTimedOut && logged_time("Solving") < iTimeLimit);
  }

  /* Check for exit request */
  if (bSolveOnly) {
    ofstream out(outFile.c_str());

    output(*solver, out);
    log_end(TOTAL_TIME);
    wrap_up();
    exit(0);
  }
  
  /* Check for load request */
  if (bLoadOnly) {
    ifstream inOut(outFile.c_str());
    ifstream inMap(mapFile.c_str());
    char check;

    inOut.get(check);
    inOut.putback(check);
    if (check == 'N') {
	 string str;
	 inOut >> str;
	 assert(str == "NO");
	 inOut >> str;
	 assert(str == "SOLUTION");

	 found_plan = false;
    }
    else {
	 solver_stmt.readSolution(inOut, inMap, solution, rsolution);
	 found_plan = true;
    }
  }

  if (!bOutputOnly && !bSolveOnly) {
    log_start("Printing", DETAIL_MINIMAL_DEBUG);
    
    if (detail_level >= DETAIL_MINIMAL_DEBUG)
	 cout << "Printing." << endl;
    
    pretty_print(found_plan, solution, rsolution);
    
    
    if (detail_level >= DETAIL_MINIMAL_DEBUG)
	 cout << "Done." << endl;
    
    log_end("Printing");
  }

  log_end(TOTAL_TIME);
  wrap_up();

  exit(0);
}

/***********************************************/
/*  Argument handling                          */
/***********************************************/

void argSetup()
{
  // Set up all the arguments:
  argParser.addArg('o', "ops-file", new StringArg("Ops file name"), true);
  argParser.addArg('I', "intermediate-file", 
			    new StringArg("file name base", "",
				   "Stores CNF/map/out file name base for read or write"), 
			    false);
  argParser.addArg('e', "explanation",
			    new SwitchArg("Turn on explanations in the LP solver", 
						   true), false);
  argParser.addArg('C', "cons-cache", new IntArg("caching policy (0-2)",
									    cache_policy, 
									    CACHE_POLICY_STRING),
			    false);
  argParser.addArg('D', "cache-depth", 
			    new IntArg("caching depth", cache_depth,
			  "With cons-cache setting, max cache depth before flushing"),
			    false);
  argParser.addArg('W', "write",
			    new SwitchArg(((string)"Write to map/cnf files") +
						   " (only matters if non-null file name base)", 
						   false), false);
  argParser.addArg('L', "load",
			    new SwitchArg(((string)"Read from map/out files") +
						   " (only matters if non-null file name base)", 
						   false), false);
  argParser.addArg('S', "solve",
			    new SwitchArg(((string)"Solve from map/cnf to out file") +
						   " (only matters if non-null file name base)", 
						   false), false);
  argParser.addArg('f', "fact-file", new StringArg("Fact file name"), true);
  argParser.addArg('T', "path", new StringArg("Common file path"), false);
  argParser.addArg('H', "restart", new IntArg("Random restart initial time",
									 0, "0 means no random restart"),
			    false);
  argParser.addArg('a', "restart-add",
			    new DoubleArg("Random restart increment", 0,
				"Random restart time additive increment (T' = m*T + a)"),
			    false);
  argParser.addArg('m', "restart-multiply",
			    new DoubleArg("Random restart multiplier", 2,
						"Random restart time multiplier (T' = m*T + a)"),
			    false);
  argParser.addArg('d', "detail", 
			    new IntArg("Detail level (0-5)", detail_level, 
						DETAIL_LEVEL_STRING), false);
  argParser.addArg('t', "time-limit", 
			    new IntArg("Time limit (seconds)", 0, 
						"0 means no time limit"), false);
  argParser.addArg('v', "cutoff", 
			    new DoubleArg("Cutoff factor for var scoring", 0.8), false);
  argParser.addArg('F', "fudge", 
			    new DoubleArg("Fudge factor on scores for var pruning", 
						   0.9), false);
  argParser.addArg('r', "cons-pref", 
			    new DoubleArg("Cons preference factor", 1, 
						   "Additive factor for constraint scores"), 
			    false);
  argParser.addArg('z', "size-factor", 
			    new DoubleArg("Size preference factor", 0, 
		   "Adjustment for constraint scores based on length (score += len*factor)"), 
			    false);
  argParser.addArg('U', "cons-unit",
			    new SwitchArg(((string)"Perform prop on constrained ") + 
						   "vars during fast unit prop?", false), false);
  argParser.addArg('G', "logic-len",
			    new IntArg("Logic cutoff length", 10, 
						"Logic learning cutoff length"), false);
//  argParser.addArg('C', "cons-len",
//			    new IntArg("Constraint cutoff length", 1, 
//						"Constraint learning cutoff length"), false);
  argParser.addArg('R', "relevance",
			    new SwitchArg("Use relevance-bounded learning", true),
			    false);
  argParser.addArg('p', "parallel",
			    new SwitchArg("Use parallel encoding", bParallel), false);
  argParser.addArg('i', "influences",
			    new SwitchArg("Use influences (only matters w/parallel)", 
						   bInfluences), false);
  argParser.addArg('c', "compact",
			    new IntArg("Compact level (0-3)", compact_level,
						COMPACT_LEVEL_STRING), false);
  argParser.addArg('x', "frame-mutex",
			    new SwitchArg("Make set infl to 0 mutex w/actual infl",
						   true), false);
  argParser.addArg('X', "random-seed",
			    new IntArg("Random # seed", -1,
						"-1 means use time"), false);
//  argParser.addArg('P', "pure",
//			    new SwitchArg("Use pure literal elimination", true),
//			    false);
  argParser.addArg('n', "naive", 
			    new SwitchArg("Naive DPLL (no learning, no CBJ)", false), 
			    false);
  argParser.addArg('l', "learn", 
			    new SwitchArg("Permit learning (no effect on CBJ)", true), 
			    false);
  argParser.addArg('N', "neg-first",
			    new SwitchArg("Always set constrained vars to false first",
						   true), false);
  argParser.addArg('s', "steps", new IntArg("# of steps in plan", -1,
								    "-1 means use hint in fact file"), 
			    false);
}

void argGather()
{
  // Pass the argument data to the appropriate place:
  strcpy(ops_file_name, 
	    ((StringArg*)argParser.getArg("ops-file"))->val.c_str());
  strcpy(fct_file_name,
	    ((StringArg*)argParser.getArg("fact-file"))->val.c_str());
  strcpy(path,
	    ((StringArg*)argParser.getArg("path"))->val.c_str());

  cnfFile = ((StringArg*)argParser.getArg('I'))->val;
  mapFile = ((StringArg*)argParser.getArg('I'))->val;
  outFile = ((StringArg*)argParser.getArg('I'))->val;

  bOutputOnly = cnfFile != "" && ((SwitchArg*)argParser.getArg("write"))->val;
  bLoadOnly = cnfFile != "" && ((SwitchArg*)argParser.getArg("load"))->val;
  bSolveOnly = cnfFile != "" && ((SwitchArg*)argParser.getArg("solve"))->val;
  cnfFile = path + cnfFile + cnfExt;
  mapFile = path + mapFile + mapExt;
  outFile = path + outFile + outExt;

  if (bOutputOnly || bLoadOnly || bSolveOnly) {
    if ((bOutputOnly && bLoadOnly) || (bOutputOnly && bSolveOnly) || 
	   (bLoadOnly && bSolveOnly)) {
	 cerr << "Must select at most one of load, write, and solve." << endl;
	 exit(1);
    }
  }

  cache_policy = ((IntArg*)argParser.getArg("cons-cache"))->val;
  cache_depth = ((IntArg*)argParser.getArg("cache-depth"))->val;

  if (cache_depth < 0 && cache_policy == CACHE_BRANCH) {
    cerr << "Error: with branch counting cache policy, "
	    << "must have non-negative cache depth." << endl;
    exit(1);
  }

  if (cache_depth <= 0 && cache_policy == CACHE_SIZE) {
    cerr << "Error: with constraint counting cache policy, "
	    << "must have positive cache depth." << endl;
    exit(1);
  }

  SATSolver::iCachePolicy = cache_policy;
  SATSolver::iCacheDepth = cache_depth;

  SATSolver::bExplanation = ((SwitchArg*)argParser.getArg("explanation"))->val;
  SATSolver::scoreCutoffFactor = 
    ((DoubleArg*)argParser.getArg("cutoff"))->val;
  SATSolver::bestScoreFudgeFactor = 
    ((DoubleArg*)argParser.getArg("fudge"))->val;
  SATSolver::constraintPreferenceFactor = 
    ((DoubleArg*)argParser.getArg("cons-pref"))->val;
  SATSolver::constraintSizeFactor = 
    ((DoubleArg*)argParser.getArg("size-factor"))->val;
  SATSolver::fastUnitPropConstrained =
    ((SwitchArg*)argParser.getArg("cons-unit"))->val;
// Cons len option has been removed (as it no longer affects anything)
  SATSolver::iConsMaxLearnSize = 1;
//    ((IntArg*)argParser.getArg("cons-len"))->val;
  SATSolver::iLogicMaxLearnSize =
    ((IntArg*)argParser.getArg("logic-len"))->val;
  SATSolver::bLearn = ((SwitchArg*)argParser.getArg("learn"))->val;
  SATSolver::bNaiveSolver = ((SwitchArg*)argParser.getArg("naive"))->val;
  SATSolver::bRelevance =
    ((SwitchArg*)argParser.getArg("relevance"))->val;
// Too obviously advantageous. Now automatically set.
  SATSolver::bUsePureLitElim = true;
//    ((SwitchArg*)argParser.getArg("pure"))->val;
  SATSolver::bConstrainedVarsNegFirst =
    ((SwitchArg*)argParser.getArg("neg-first"))->val;
  bParallel = ((SwitchArg*)argParser.getArg("parallel"))->val;
  bInfluences = ((SwitchArg*)argParser.getArg("influences"))->val;
  bMutexInfl = ((SwitchArg*)argParser.getArg('x'))->val;
  detail_level = ((IntArg*)argParser.getArg("detail"))->val;
  compact_level = ((IntArg*)argParser.getArg("compact"))->val;
  if (((IntArg*)argParser.getArg("time-limit"))->val == 0) {
    bTimeLimit = false;
    iTimeLimit = INT_MAX;
  }
  else {
    bTimeLimit = true;
    iTimeLimit = ((IntArg*)argParser.getArg("time-limit"))->val * 100;
  }
  if (((IntArg*)argParser.getArg("restart"))->val == 0) {
    bRandomRestart = false;
    iInitLimit = iTimeLimit;
  }
  else {
    bRandomRestart = true;
    dRRMultiplier = ((DoubleArg*)argParser.getArg('m'))->val;
    dRRIncrement = ((DoubleArg*)argParser.getArg('a'))->val;
    iInitLimit = ((IntArg*)argParser.getArg("restart"))->val * 100;
  }

  SATSolver::bTimeLimit = bTimeLimit || bRandomRestart;

  randomSeed = ((IntArg*)argParser.getArg('X'))->val;
  if (randomSeed == -1) {
    randomSeed = time(NULL);
  }
}

void argPrint()
{
  // Output a summary of the arguments
  cout << "Operator file name: " << ops_file_name << endl;
  cout << "Fact file name: " << fct_file_name << endl;
  cout << "Path: " << path << endl;
  if (bOutputOnly) {
    cout << "Outputs to: " << cnfFile << " and " << mapFile << ". Then, exits." << endl;
  }
  if (bSolveOnly) {
    cout << "Solves from: " << cnfFile << " and " << mapFile << "." << endl;
    cout << "  Then, outputs to " << outFile << ", and exits." << endl;
  }
  if (bLoadOnly) {
    cout << "Loads from: " << cnfFile << ", " << mapFile << ", and " << outFile << "."
	    << endl << "  Outputs solution. Then, exits." << endl;
  }
  cout << "Score cutoff factor: " << SATSolver::scoreCutoffFactor << endl;
  cout << "Score fudge factor: " << SATSolver::bestScoreFudgeFactor << endl;
  cout << "Constraint score factor: " <<
    SATSolver::constraintPreferenceFactor << endl;
  cout << (SATSolver::fastUnitPropConstrained ? 
		 "Performs " : "Does not perform ")
	  << "fast unit propagation on constrained variables." << endl;
  if (SATSolver::bNaiveSolver || !SATSolver::bLearn) {
    cout << "Does not use learning." << endl;
    cout << (SATSolver::bNaiveSolver ? "Does not" : "Does")
	    << " perform CBJ." << endl;
  }
  else {
    cout << "Performs learning." << endl << "Performs CBJ." << endl;
    cout << "Constraint learn order: " << SATSolver::iConsMaxLearnSize 
	    << endl;
    cout << "Logic learn order: " << SATSolver::iLogicMaxLearnSize << endl;
    cout << (SATSolver::bRelevance ? "Does " : "Does not ")
	    << "perform relevance bounded learning." << endl;
  }
  cout << (SATSolver::bUsePureLitElim ? "Does " : "Does not ")
	  << "perform pure literal elimination." << endl;
  cout 
    << (SATSolver::bConstrainedVarsNegFirst ? "Forces " : "Does not force ")
    << "constrained vars to take the value false first." << endl;
  cout << "Uses " << (bParallel ? "parallel" : "serial") << " encoding."
	  << endl;
  cout << (bInfluences ? "Uses " : "Does not use ") << " influences."
	  << endl;
  cout << (bMutexInfl ? "Adds " : "Does not add ") 
	  << " mutex influence clause." << endl;
  cout << "Random seed: " << randomSeed << endl;
  cout << "Steps: " << ((IntArg*)argParser.getArg("steps"))->val << endl;
  if (bTimeLimit) {
    cout << "Runs under time limit: " << iTimeLimit / 100.0 << endl;
  }
  else {
    cout << "Does not run under time limit" << endl;
  }
  if (bRandomRestart) {
    cout << "Uses random restart commencing at: " << iInitLimit / 100.0 
	    << endl;
    cout << "Restart multiplier: " << dRRMultiplier << endl;
    cout << "Restart addend: " << dRRIncrement << endl;
  }
  else {
    cout << "Does not use random restart." << endl;
  }
  cout << "Compact level: " << compact_level << endl;
  cout << "Detail level: " << detail_level << endl;
}

