// !!! Problems: 
// !!!   Need to move indexing out of exc finding (to catch _all_ exes)

/*
 * This file contains functions used to construct the SAT encoding of 
 * the planning problem from the fully instantiated PDDL/ipp version
 */

#include "ipp.h"
#include "cnf_stmt.h"

#include <assert.h>
#include <hash_set>
#include <hash_map>

fact_list find_preconds(fact_list clause);

/* mapping from instantiated predicates to actions/clauses containing
   those predicates */
#ifndef HASH_TOKEN_LIST
#define HASH_TOKEN_LIST
struct equal_to<token_list>
{
  bool operator()(const token_list & a, const token_list & b) {
    return token_lists_equal(a, b);
  }
};

struct hash<token_list>
{
  size_t operator()(const token_list & hashee) const {
    token_list t = hashee;
    size_t result = 0;
    while (t != NULL) {
	 token s = t->item;
	 while (*s) {
	   // Relies on 32 bit size_t.
	   result = (result << 1) + (result >> 31);
	   result += (size_t)*s;
	   s++;
	 }
	 t = t->next;
    }

    return result;
  }
};
#endif

struct ActionClause {
  operator_list action;
  fact_list clause;
  ActionClause(operator_list a = NULL, fact_list c = NULL) : 
    action(a), clause(c) {}
};

struct ActionClauseList {
  ActionClause * head;
  ActionClauseList * tail;
  ActionClauseList(ActionClause * h, ActionClauseList * t) {
    head = h;
    tail = t;
  }
};

typedef hash_map<token_list , ActionClauseList * > PredActionMap;

PredActionMap predActionMap;

ActionClauseList * find_term(token_list term)
{
  return predActionMap[term];
}

void index_term(operator_list action, fact_list clause, token_list term)
{
  ActionClauseList * acl = predActionMap[term];
  
  predActionMap[term] = new ActionClauseList(new ActionClause(action, clause),
									acl);
}

void index_clause(operator_list action, fact_list clause)
{
  fact_list f = clause;
  fact_list new_clause;
  bool precond = false, effect = false;
  bool conditional;

  while (f != NULL && get_adl_token(f) != RBRACK_CONST) {
    if (is_effect(f))
	 effect = true;
    if (is_precond(f))
	 precond = true;
    f = f->next;
  }
  
  if (effect && precond) {
    conditional = true;
    new_clause = find_preconds(clause);
  }
  else {
    conditional = false;
    new_clause = NULL;
  }

  f = clause;
  while (f != NULL && get_adl_token(f) != RBRACK_CONST) {
    token_list term;
    
    /* Need to negate the term if it's a precondition before indexing it */
    if (conditional && is_precond(f)) {
	 term = new_token_list();
	 term->item = new_token(strlen(f->item->item) + 2);
	 if (is_negated(f)) {
	   strcpy(term->item, f->item->item + 1);
	 }
	 else {
	   sprintf(term->item, "%s%s", "!", f->item->item);
	 }
	 term->next = f->item->next;
    }
    else {
	 term = f->item;
    }
    
    if (get_adl_token(f) == LIT_CONST) {
	 index_term(action, new_clause, term);
    }
    f = f->next;
  }
}

void index_action(operator_list action, fact_list clauses)
{
  /* Step through the list of clauses breaking them up and sending
	them to index_clause. Note that this munges the memory in clauses! */
  fact_list current = clauses, temp;
  while (clauses != NULL) {
    current = clauses;
    clauses = clauses->next;
    while (get_adl_token(clauses) != RBRACK_CONST) {
	 clauses = clauses->next;
    }

    temp = clauses->next;
    clauses->next = NULL;
    index_clause(action, current);
    clauses = temp;
  }
}

/* function prototypes */
void build_inits();
void build_actions();
void build_exclusivity();
void build_parallel_exclusivity();
void build_serial_exclusivity();
void build_frame_axioms();
void build_constraint_frame_axioms();
void build_predicate_frame_axioms();
void build_goal();
void build_action(operator_list i_op);

#define PARALLEL_ACTIONS // uncomment to enable parallel actions

#define TIME_DIGITS 4

#ifndef HASH_PAIR_STRING_INT
#define HASH_PAIR_STRING_INT
struct hash<pair<string, int> >
{
  size_t operator()(const pair<string,int> & p) const {
    unsigned int i;
    size_t result = 0;
    for (i = 0; i < p.first.length(); i++) {
	 result <<= 1;
	 result += (size_t)p.first[i];
    }

    result <<= 1;
    result += p.second;

    return result;
  }
};
#endif

hash_set<pair<string, int> > completedInfluences;

void time_label_fact_list(fact_list fl, int step);

/* Labels all variables needing relabeling according to
   the time provided. Variables that appear
   as fluents in the domain_variables list must be renumbered */
/* ASSUMES POSITIVE TOKEN_LIST (no '!') */
void time_label_constraint(token_list tl, int step, int adjusted_step)
{
  token_list type;
  BOOLEAN set_constraint = FALSE;
  BOOLEAN bPrimed;

  /* INFLUENCEs no longer need to be separate 
	from SETs and if influences are on, they don't need 
	the changed value on the right and they need a new
	clause restricting their value */
  if (strcmp(tl->item, INFLUENCE_STR) == SAME) {
    token t = new_token(strlen(SET_STR) + 1);
    strcpy(t, SET_STR);
    my_free(tl->item);
    tl->item = t;
    
    if (bInfluences) {
	 /* !!! Memory leak (fixable) !!! */
	 tl->next->next->next = tl->next->next->next->next->next;

	 /* If it hasn't been already, construct and enter a clause
	    setting this influence to zero if it isn't active */
	 if (completedInfluences.find(pair<string,int>(tl->next->item,step)) == 
		completedInfluences.end()) {
	   /* start with -a v -b and then a v b */
	   token_list effect;
	   token_list end, temp;
	   fact_list fl = make_adl_fact(OR_CONST);
	   fl->next = new_fact_list();
	   
	   fl->next->item = copy_complete_token_list(tl, &end);

	   /* prepend the necessary EFFECT_STR */
	   effect = new_token_list();
	   effect->item = new_token(strlen(EFFECT_STR) + 1 + 1);
	   sprintf(effect->item, "!%s", EFFECT_STR);
	   effect->next = fl->next->item;
	   fl->next->item = effect;
	   
	   /* store the right side of the equation then copy up to that
		 side */
	   temp = tl->next->next->next;
	   tl->next->next->next = NULL;
	   fl->next->next = new_fact_list();
	   fl->next->next->item = copy_complete_token_list(tl, &end);

	   /* restore the equation */
	   tl->next->next->next = temp;

	   /* tack on the zero for the right side */
	   end->next = new_token_list();
	   end->next->item = new_token(strlen("0") + 1);
	   strcpy(end->next->item, "0");

	   /* prepend the necessary EFFECT_STR */
	   effect = new_token_list();
	   effect->item = new_token(strlen(EFFECT_STR) + 1 + 1);
	   sprintf(effect->item, "!%s", EFFECT_STR);
	   effect->next = fl->next->next->item;
	   fl->next->next->item = effect;

	   /* polish off the OR clause */
	   fl->next->next->next = make_adl_fact(RBRACK_CONST);

	   time_label_fact_list(fl, step);

	   /* May not want the negative half */
	   if (bMutexInfl) {
		/* send the whole thing off to get added */
		solver_stmt.add(fl);
	   }

	   /* Now, add set var' = var + infl v set var' = 0 */
	   fl->next->item->item++;
	   fl->next->next->item->item++;

	   solver_stmt.add(fl);

	   /* Now free up the leftovers */
	   fl->next->item->item--;
	   fl->next->next->item->item--;

	   free_complete_fact_list(fl);

	   /* Note that this has been handled */
	   completedInfluences.insert(pair<string,int>(tl->next->item,step));
	 }
    }
  }

  /* to handle the fact that all but the first var
	in a SET_STR must occur at PRECOND time. */
  if (strcmp(tl->item, SET_STR) == SAME) {
    set_constraint = TRUE;
    tl = tl->next;
  }

  while (tl) {
    if (tl->item[0] == '?') {
	 if (is_primed_prune(tl)) {
	   adjusted_step++;
	   bPrimed = true;
	 }
	 else {
	   bPrimed = false;
	 }
	 /* if the variable appears as a fluent in the 
	    domain_variables list, time label it */
	 if ((type = type_already_known(tl->item, domain_vars_list)) &&
		 strcmp(type->item, FLUENT_STR) == SAME) {
	   token t = new_token(strlen(tl->item) + strlen(TIME_PREFIX) + 
					   strlen(VAR_INTERNAL_CONNECTOR) + 
					   TIME_DIGITS + 1);
	   sprintf(t, "%s%s%d%s%s", "?", TIME_PREFIX, adjusted_step,
			                  VAR_INTERNAL_CONNECTOR, tl->item + 1);
	   my_free(tl->item);
	   tl->item = t;
	 }
	 
	 /* if the variable is a parameter or temporary influence variable
	    (one that has been numbered) time-label it independent of 
	    precond/effect status */
	 if (tl->item[1] >= '0' && tl->item[1] <= '9') {
	   token t = new_token(strlen(tl->item) + strlen(TIME_PREFIX) + 
					   strlen(VAR_INTERNAL_CONNECTOR) + 
					   TIME_DIGITS + 1);
	   sprintf(t, "%s%s%d%s%s", "?", TIME_PREFIX, step,
			                  VAR_INTERNAL_CONNECTOR, tl->item + 1);
	   my_free(tl->item);
	   tl->item = t;
	 }
	 if (bPrimed) {
	   adjusted_step--;
	 }
    }

    if (set_constraint) {
	 /* the first var of the set has been taken care of;
	    pretend that the rest is a test: */
	 set_constraint = FALSE;
	 adjusted_step--;
    }
    tl = tl->next;
  }
}

/* Labels all the predicates in this clause according
   to whether they are PRECONDS (step i), EFFECTS (step i+1),
   or neither (step i) */
void time_label_fact_list(fact_list fl, int step)
{
  token time_label;
  int actual_step;     /* the step to label this token_list at */
  BOOLEAN negated;     /* was this originally negated? */
  token_list tl;       /* the token_list that will finally follow the 
					 time label */
  BOOLEAN constraint_form;

  while (fl) {
    /* skip non LITs */
    if (get_adl_token(fl) != LIT_CONST) {
	 fl = fl->next;
	 continue;
    }

    negated = fl->item->item[0] == '!';
    constraint_form = is_constraint_form(fl);

    /* effects */
    if (negated ? 
	   (strcmp(fl->item->item + 1, EFFECT_STR) == SAME) : 
	   (strcmp(fl->item->item, EFFECT_STR) == SAME)) {
	 actual_step = step + 1;
	 tl = fl->item->next;
	 fl->item->next = NULL;
	 free_complete_token_list(fl->item);
    }
    /* preconds */
    else if (negated ? 
		   (strcmp(fl->item->item + 1, PRECOND_STR) == SAME) : 
		   (strcmp(fl->item->item, PRECOND_STR) == SAME)) {
	 actual_step = step;
	 tl = fl->item->next;
	 fl->item->next = NULL;
	 free_complete_token_list(fl->item);
    }
    else {
	 actual_step = step;
	 if (negated) {
	   token new_first = new_token(strlen(fl->item->item));
	   strcpy(new_first, fl->item->item + 1);
	   my_free(fl->item->item);
	   fl->item->item = new_first;
	 }
	 tl = fl->item;
    }

    if (strcmp(tl->item, EQ_STR) == SAME) {
	 if (negated) {
	   token neg = new_token(strlen(tl->item) + 1 + 1);
	   sprintf(neg, "%s%s", "!", tl->item);
	   my_free(tl->item);
	   tl->item = neg;
	 }
	 fl->item = tl;
    }
    else if (constraint_form) {
	 time_label_constraint(tl, step, actual_step);
	 if (negated) {
	   token neg = new_token(strlen(tl->item) + 1 + 1);
	   sprintf(neg, "%s%s", "!", tl->item);
	   my_free(tl->item);
	   tl->item = neg;
	 }
	 fl->item = tl;
    }
    else {
	 time_label = new_token(strlen(TIME_PREFIX) + TIME_DIGITS + 1 + 1);
	 sprintf(time_label, "%s%s%d", negated ? "!" : "", TIME_PREFIX, 
		    actual_step);
	 fl->item = new_token_list();
	 fl->item->item = time_label;
	 fl->item->next = tl;
    }

    fl = fl->next;
  }
}


/* take the ops, inits, and so forth and transform them into 
   cnf form clauses, adding them to solver_stmt */
void build_cnf_stmt()
{
  if (detail_level >= DETAIL_MINIMAL_DEBUG)
    cerr << "Starting init build" << endl;

  log_start("Build Inits", DETAIL_STATS);
  build_inits();
  log_end("Build Inits");

  if (detail_level >= DETAIL_MINIMAL_DEBUG)
    cerr << "Starting goal build" << endl;

  log_start("Build Goal", DETAIL_STATS);
  build_goal();
  log_end("Build Goal");

  if (detail_level >= DETAIL_MINIMAL_DEBUG)
    cerr << "Starting frame axiom build" << endl;

  log_start("Build Frame Axioms", DETAIL_STATS);
  build_frame_axioms();
  log_end("Build Frame Axioms");

  if (detail_level >= DETAIL_MINIMAL_DEBUG)
    cerr << "Starting action build" << endl;

  log_start("Build Actions", DETAIL_STATS);
  build_actions();
  log_end("Build Actions");

  if (detail_level >= DETAIL_MINIMAL_DEBUG)
    cerr << "Starting exclusivity build" << endl;
  log_start("Build Exclusions", DETAIL_STATS);
  build_exclusivity(); 
  log_end("Build Exclusions");
}

/*
 * need to both construct unit clauses for the given initial facts and
 * construct unit clauses for all facts implied by closed word assumption
 */
void build_inits()
{
  fact_list fl, temp, clause;

  /* label all the initial facts as happening at time -1!! Remember,
	they're EFFECTS, so set the up as happening before everything starts
	(time_label will fix up the timing */

  time_label_fact_list(orig_initial_facts, -1);

  /* add orig_initial_facts to the solver_stmt, but as OR clauses! */

  fl = orig_initial_facts;
  while (fl) {
    temp = fl;
    fl = fl->next;

    clause = make_adl_fact(OR_CONST);
    clause->next = temp;
    temp->next = make_adl_fact(RBRACK_CONST);
    solver_stmt.add(clause);
  }
}

/*
 * Add actions in the form action -> preconditions and effects to the
 * solver_stmt.
 */
/*
 * I will assume here that I have in "operators" all the instantiated
 * operators with operators->hooks holding all the preconditions/effects.
 * moreover, I assume that the hooks are in the form of fact_lists as
 * list of clauses. each clause is a disjunction with each token/disjunct
 * a CONNECTED string of predicate/arguments.
 */
void build_actions()
{
  operator_list i_op;

  /* an action clause for each instantiation of an operator */
  for (i_op = operators; i_op; i_op = i_op->next) {
    build_action(i_op);
  }
}

/* for each clause with a positive influence as a result, negate the
   non-effect tokens and or it with the negation of the influence; in
   other words, influences are iff with their actions. */
/* !! I assume each clause has at most one effect! This stems from the
   fact that I do not support disjunctive effects !! */
void iff_influences(token negated_action, fact_list * fl)
{
  fact_list clause = *fl;
  fact_list temp;
  fact_list current;
  token_list infl;
  token_list end;

  while (clause) {
    assert(get_adl_token(clause) == OR_CONST);

    current = clause->next;

    infl = NULL;
    while (get_adl_token(current) != RBRACK_CONST) {
	 if (!is_negated(current) && is_influence(current)) {
	   infl = current->item;
	 }

	 current = current->next;
    }

    if (infl) {
	 /* start with the OR */
	 fact_list iff_clause = make_adl_fact(OR_CONST);

	 /* add the influence */
	 iff_clause->next = new_fact_list();
	 iff_clause->next->item = copy_complete_token_list(infl, &end);

	 /* add the negated action */
	 iff_clause->next->next = new_fact_list();
	 iff_clause->next->next->item = new_token_list();
	 iff_clause->next->next->item->item = new_token(strlen(negated_action)+1);
	 strcpy(iff_clause->next->next->item->item, negated_action);

	 /* now, add the preconds */
	 iff_clause->next->next->next = find_preconds(clause);

	 /* now, cap off with an RBRACK */
	 for (temp = iff_clause; temp->next; temp = temp->next)
	   ;
	 temp->next = make_adl_fact(RBRACK_CONST);
	 /*!!!*/
	 /* now, negate the work so far */
	 negate_fact_list(iff_clause);

	 /* tack this clause onto the beginning of the fl */
	 temp->next->next = *fl;
	 *fl = iff_clause;

    }

    clause = current->next;
  }
}

/* add one particular action at each time step into the clause list */
/* action -> preconditions & effects */
/* !action v (preconditions & effects) */
void build_action(operator_list i_op)
{
  fact_list action_clause, negated_action, end;
  token negated_action_name;

  /* make the action token */
  /* negated_action will get munged before I reach the last
	time I need the negated action name, so make two copies */
  negated_action = new_fact_list();
  negated_action->item = new_token_list();
  negated_action_name = new_token(strlen(i_op->name) + 1 + 1);
  sprintf(negated_action_name, "%s%s", "!", i_op->name);
  negated_action->item->item = new_token(strlen(negated_action_name) + 1);
  strcpy(negated_action->item->item, negated_action_name);

  /* start off the OR clause */
  action_clause = make_adl_fact(OR_CONST);

  /* add the action token */
  action_clause->next = negated_action;
  end = action_clause->next;

  if (i_op->hooks) {
	 /* add all preconds and effects. Note that the AND is necessary since
	    top-level (hooks) is implicitly ANDed, but at this level it needs to
	    be explicit */
	 negated_action->next = make_adl_fact(AND_CONST);
	 negated_action->next->next = copy_complete_fact_list(i_op->hooks, &end);
	 end->next = make_adl_fact(RBRACK_CONST);
	 end = end->next;
  }

  end->next = make_adl_fact(RBRACK_CONST);

  /* put into CNF form */
  action_clause = rewrite_in_cnf(&action_clause);
  
  /* If influences are on, need to activate a partial influence
	variable iff the action causes it */
  if (bInfluences)
    iff_influences(negated_action_name, &action_clause);

  if (!action_clause)
    return;
    
  /* for each step, construct the action clauses */
  for (int i = 0; i < gproblem_length; i++) {
    fact_list dummy;
    fact_list f = copy_complete_fact_list(action_clause, &dummy);

    time_label_fact_list(f, i);
    
    solver_stmt.add(f);

    free_complete_fact_list(f);
  }

  /* Finally, index this action by its preconditions/effects. Note
     that index_action uses the memory of action_clause */
  index_action(i_op, action_clause);  
  my_free(negated_action_name);
}

void build_exclusivity()
{
  if (bParallel)
    build_parallel_exclusivity();
  else
    build_serial_exclusivity();
}

bool are_clauses_cons_exclusive(fact_list clause1, fact_list clause2)
{
  fact_list curr1, curr2;
  token var;

  /* constraints are exclusive for now iff the change the same
	var and at least one is a set !!! */
  for (curr1 = clause1->next; 
	  get_adl_token(curr1) != RBRACK_CONST; 
	  curr1 = curr1->next) {
    if (!is_negated(curr1) && is_effect(curr1) &&
	   is_constraint_form(curr1)) {
	 /* skip the EFFECT and the SET to get at the var */
	 var = curr1->item->next->next->item;

	 for (curr2 = clause2->next; 
		 get_adl_token(curr2) != RBRACK_CONST;
		 curr2 = curr2->next) {
	   if (!is_negated(curr2) && is_effect(curr2) &&
		  is_constraint_form(curr2)) {
		token t = curr2->item->next->next->item;

		if (strcmp(t, var) == SAME) {
		  return true;
		}
	   }
	 }
    }
  }

  return false;
}

bool are_clauses_pred_exclusive(fact_list clause1, fact_list clause2)
{
  fact_list curr1, curr2;
  static token_list tl = new_token_list();
  static token first_tok = 
    new_token(strlen(EFFECT_STR)+strlen(PRECOND_STR)+1+1);
  bool is_effect1 = false, is_effect2 = false;
  bool is_conflict = false;

  /* pred exclusive if: one clause deletes the preconds of the other,
	either conditional or unconditional. That means search around for
	an effect in each clause to see whether preconds are conditional
	first, then decide what to do. */

  for (curr1 = clause1->next;
	  get_adl_token(curr1) != RBRACK_CONST;
	  curr1 = curr1->next) {
    if (is_effect(curr1)) {
	 is_effect1 = true;
	 break;
    }
  }
  
  for (curr2 = clause2->next;
	  get_adl_token(curr2) != RBRACK_CONST;
	  curr2 = curr2->next) {
    if (is_effect(curr2)) {
	 is_effect2 = true;
	 break;
    }
  }

  /* preconditions don't exclude */
  if (!is_effect1 && !is_effect2)
    return false;

  for (curr1 = clause1->next; 
	  get_adl_token(curr1) != RBRACK_CONST; 
	  curr1 = curr1->next) {
    if (is_constraint_form(curr1))
	 continue;
    
    bool is_this_effect = is_effect(curr1);
    /* if this is an effect, then we want to negate it unless 
	  clause2 contains conditional preconds (which are already negated).
	  if this isn't an effect, then we want to negate it unless
	  clause1 contains condition preconds (which are already negated). */
    bool negate_tok = is_this_effect ? !is_effect2 : !is_effect1;

    if (is_negated(curr1))
	 sprintf(first_tok, "%s%s", negate_tok ? "" : "!", 
		    is_this_effect ? PRECOND_STR : EFFECT_STR);
    else
	 sprintf(first_tok, "%s%s", negate_tok ? "!" : "", 
		    is_this_effect ? PRECOND_STR : EFFECT_STR);	 

    tl->item = first_tok;
    tl->next = curr1->item->next;

    if (find_token_list_in_clause(clause2, tl)) {
	 is_conflict = true;
	 break;
    }
  }

  return is_conflict;
}

/* finds exclusivity based on either constraints or predicates */
bool are_clauses_exclusive(fact_list clause1, fact_list clause2)
{
  return (are_clauses_cons_exclusive(clause1, clause2) ||
		are_clauses_pred_exclusive(clause1, clause2));
}

/* actually construct and add the exclusion for each time step */
/* assumes that both clauses come in in the appropriate form to be put
   into the exclusion (stripped of effects and (if not conditional),
   empty */
void build_parallel_exclusion(operator_list op1, fact_list clause1, 
						operator_list op2, fact_list clause2)
{
  fact_list ex, tmp, end;
  int i;

  /* If the two actions are the same, don't build an exclusion
	between them! */
  if (op1 == op2)
    return;

  /* construct the exclusion clause */
  ex = tmp = make_adl_fact(OR_CONST);

  /* make the action1 token */
  tmp->next = new_fact_list();
  tmp = tmp->next;
  tmp->item = new_token_list();
  tmp->item->item = new_token(strlen(op1->name) + 1 + 1);
  strcpy(tmp->item->item, "!");
  strcat(tmp->item->item, op1->name);
  
  /* add on the preconds of op1 */
  tmp->next = clause1;
  for ( ; tmp->next; tmp = tmp->next) {
    ;
  }

  /* add on op2's action token */
  tmp->next = new_fact_list();
  tmp = tmp->next;
  tmp->item = new_token_list();
  tmp->item->item = new_token(strlen(op2->name) + 1 + 1);
  strcpy(tmp->item->item, "!");
  strcat(tmp->item->item, op2->name);

  /* add on op2's conds */
  tmp->next = clause2;
  for ( ; tmp->next; tmp = tmp->next) {
    ;
  }

  /* finish the clause */
  tmp->next = make_adl_fact(RBRACK_CONST);

  /* step through time, prepending onto solver_stmt */
  for (i = 0; i < gproblem_length; i++) {
    tmp = copy_complete_fact_list(ex, &end);
    time_label_fact_list(tmp, i);
    
    end->next = NULL;
    solver_stmt.add(tmp);
    free_complete_fact_list(tmp);
  }

  free_complete_fact_list(ex);
}

void build_pred_term_exclusivity(operator_list op, fact_list clause, 
						   fact_list stripped_clause, 
						   fact_list term)
{
  /* First, build the precondition to search for; the negation
	of term */
  bool negated = is_negated(term);
  token_list precond = new_token_list();
  
  precond->item = new_token(strlen(PRECOND_STR) + 1 + 1);
  sprintf(precond->item, "%s%s", negated ? "" : "!", PRECOND_STR);
  precond->next = term->item->next;

  /* Next, find all the clauses associated with this precondition */
  ActionClauseList * action_clauses = find_term(precond);

  while (action_clauses != NULL) {
    operator_list op2 = action_clauses->head->action;
    fact_list clause2 = action_clauses->head->clause;
    action_clauses = action_clauses->tail;

    build_parallel_exclusion(op, stripped_clause, op2, clause2);
  }
}

void build_cons_term_exclusivity(operator_list op, fact_list clause, 
						   fact_list stripped_clause, 
						   fact_list term)
{
  /* First, build the effect to search for; another SET/INFLUENCE
     statement */
  token_list goal = new_token_list();
  
  goal->item = term->item->item;
  goal->next = new_token_list();
  goal->next->item = new_token(strlen(SET_STR) + 1);
  strcpy(goal->next->item, SET_STR);
  goal->next->next = term->item->next->next;

  /* Next, find all the clauses associated with this effect */
  ActionClauseList * action_clauses = find_term(goal);

  while (action_clauses != NULL) {
    operator_list op2 = action_clauses->head->action;
    fact_list clause2 = action_clauses->head->clause;
    action_clauses = action_clauses->tail;

    build_parallel_exclusion(op, stripped_clause, op2, clause2);
  }

  /* if the term is a SET statement. repeat the same process for influence
	statements. */
  if (is_set(term)) {
    token_list goal = new_token_list();
    
    goal->item = term->item->item;
    goal->next = new_token_list();
    goal->next->item = new_token(strlen(INFLUENCE_STR) + 1);
    strcpy(goal->next->item, INFLUENCE_STR);
    goal->next->next = term->item->next->next;
    
    /* Next, find all the clauses associated with this effect */
    ActionClauseList * action_clauses = find_term(goal);
    
    while (action_clauses != NULL) {
	 operator_list op2 = action_clauses->head->action;
	 fact_list clause2 = action_clauses->head->clause;
	 action_clauses = action_clauses->tail;
	 
	 build_parallel_exclusion(op, clause, op2, clause2);
    }
  }
}
  
/* build exclusivity for a single clause */
void build_clause_exclusivity(operator_list op1, fact_list clause1)
{
  /* construct a stripped version of the clause that contains only the
	information important for exclusions (nothing unless conditional,
	and in that case, just the preconds) */
  fact_list new_clause = NULL, f = clause1;
  bool precond = false, effect = false;

  while (f != NULL && get_adl_token(f) != RBRACK_CONST) {
    if (is_effect(f))
	 effect = true;
    if (is_precond(f))
	 precond = true;
    f = f->next;
  }
  
  if (effect && precond) {
    new_clause = find_preconds(clause1);
  }
  else {
    new_clause = NULL;
  }


  /* find all the clauses which are exclusive with this one and build
     exclusion clauses for each. */
  for (fact_list f = clause1->next; get_adl_token(f) != RBRACK_CONST;
	  f = f->next) {
    if (is_effect(f)) {
	 if (is_constraint_form(f)) {
	   /* build constraint exclusions based on this effect. */
	   build_cons_term_exclusivity(op1, clause1, new_clause, f);
	 }
	 else {
	   /* build predicate exclusions based on this effect. */
	   build_pred_term_exclusivity(op1, clause1, new_clause, f);
	 }
    }
  }
}

void build_parallel_exclusivity(operator_list op1)
{
  fact_list clause1;
  fact_list curr_fact;

  for (clause1 = op1->hooks; clause1; clause1 = clause1->next) {
    assert(get_adl_token(clause1) == OR_CONST);

    build_clause_exclusivity(op1, clause1);

    for (curr_fact = clause1; get_adl_token(curr_fact) != RBRACK_CONST;
	    curr_fact = curr_fact->next) {
	 ;
    }

    clause1 = curr_fact;

    assert(get_adl_token(clause1) == RBRACK_CONST);
  }
}

void build_parallel_exclusivity()
{
  int count = 0;
  operator_list i_op;

  for (i_op = operators; i_op; i_op = i_op->next)
    count++;

  for (i_op = operators; i_op; i_op = i_op->next) {
    build_parallel_exclusivity(i_op);

    if (detail_level >= DETAIL_MINIMAL_DEBUG)
	 cerr << count-- << " ";
  }
  if (detail_level >= DETAIL_MINIMAL_DEBUG)
    cerr << endl;
}


void build_serial_exclusion(operator_list op1, operator_list op2, 
					   int time_step)
{
  fact_list dummy;
  token not_act1, not_act2;
  token_list tl;

  dummy = make_adl_fact(OR_CONST);
	   
  not_act1 = new_token(strlen(op1->name) + 1 + 1);
  sprintf(not_act1, "%s%s", "!", op1->name);
	   
  not_act2 = new_token(strlen(op2->name) + 1 + 1);
  sprintf(not_act2, "%s%s", "!", op2->name);
	   
  dummy->next = new_fact_list();
  tl = new_token_list();
  tl->item = not_act1;
  dummy->next->item = tl;
	   
  dummy->next->next = new_fact_list();
  tl = new_token_list();
  tl->item = not_act2;
  dummy->next->next->item = tl;
	   
  dummy->next->next->next = make_adl_fact(RBRACK_CONST);

  time_label_fact_list(dummy, time_step);

  solver_stmt.add(dummy);
  free_complete_fact_list(dummy);
}

/* put in pairwise exclusivity clauses */
void build_serial_exclusivity()
{
  operator_list op1, op2;
  int i;

  for (op1 = operators; op1; op1 = op1->next) {
    for (op2 = op1->next; op2; op2 = op2->next) {
	 for (i = 0; i < gproblem_length; i++) {
	   build_serial_exclusion(op1, op2, i);
	 }
    }
  }
}

/*
 * Makes a fact_list consisting of copies of each PRECOND token_list
 * in the given clause (so up to RBRACK) (NULL if no PRECOND token_lists)
 */
fact_list find_preconds(fact_list clause)
{
  int adl_token;
  fact_list result = NULL, temp;
  token_list dummy;

  if (!clause)
    return NULL;

  while ((adl_token = get_adl_token(clause)) != RBRACK_CONST) {
    if (adl_token == LIT_CONST) {
	 if ((clause->item->item[0] != '!' &&
		 strcmp(clause->item->item, PRECOND_STR) == SAME) ||
		(clause->item->item[0] == '!' &&
		 strcmp(clause->item->item + 1, PRECOND_STR) == SAME)) {
	   temp = new_fact_list();
	   temp->item = copy_complete_token_list(clause->item, &dummy);
	   temp->next = result;
	   result = temp;
	 }
    }

    clause = clause->next;
  }

  return result;
}

/*
 * Return true iff left == right or one is a SET and the other is
 * a SET or INFLUENCE both with the same object.
 */
BOOLEAN loose_token_lists_equal(token_list left, token_list right)
{
  if (!left || !left->item || !left->next || !left->next->item ||
	 !left->next->next || !left->next->next->item ||
	 !right || !right->item || !right->next || !right->next->item ||
	 !right->next->next || !right->next->next->item) {
    return token_lists_equal(left, right);
  }

  if (strcmp(left->item, EFFECT_STR) == SAME &&
	 strcmp(right->item, EFFECT_STR) == SAME) {
    token right_changed_tok;
    token left_changed_tok;

    BOOLEAN right_set, right_change;
    BOOLEAN left_set, left_change;

    right_set = strcmp(right->next->item, SET_STR) == SAME;
    right_change = right_set || 
	 (strcmp(right->next->item, INFLUENCE_STR) == SAME);

    left_set = strcmp(left->next->item, SET_STR) == SAME;
    left_change = left_set || 
	 (strcmp(left->next->item, INFLUENCE_STR) == SAME);

    /* If both are changes (INFLUENCEs or SETs) and at least one
	  is a SET _or_ influences are off, they are judged by what 
	  they change */
    if ((left_set || right_set || !bInfluences) && 
	   (left_change && right_change)) {
    
	 /* Need to find the changed variable. For a SET that's just the
	    left side of the equal. For an INFLUENCE, it's both the left
	    and right side of the equal, but the one on the left has been
	    renumbered, so grab the one on the right. */
	 right_changed_tok = 
	   (right_set ?
	    (right->next->next->item) :
	    (right->next->next->next->next->item));

	 left_changed_tok = 
	   (left_set ?
	    (left->next->next->item) :
	    (left->next->next->next->next->item));

	 return strcmp(left_changed_tok, right_changed_tok) == SAME;
    }
  }

  return token_lists_equal(left, right);
}

/*
 * frame axiom helper - if effect is in op->hooks, add the appropriate
 * frame axiom baggage to end and update end
 */
/* assumes cnf form (list of OR clauses) */
void try_add_action_to_frame(operator_list op, token_list effect, 
					    fact_list * end)
{
  fact_list clause, current, conds;
  fact_list temp, and_end;

  assert(end && *end);

  /* check for effect in each OR clause of op->hooks */
  for (clause = op->hooks; clause; clause = clause->next) {

    /* check each element of the clause for the effect */
    assert(get_adl_token(clause) == OR_CONST);
    for (current = clause->next; 
	    current && get_adl_token(current) != RBRACK_CONST;
	    current = current->next) {
	 if (loose_token_lists_equal(current->item, effect)) {
	   /* this clause contains the effect sought; build the PRECOND components
		 of the clause up */
	   conds = find_preconds(clause);
	   negate_fact_list(conds);
	   /* note that I do have to negate the preconditions because
		 they have already been negated once
		 (!preconds v effect == preconds -> effect) */

	   if (!conds) {
		/* if conds is NULL, only the action is necessary */
		conds = new_fact_list();
		conds->item = new_token_list();
		conds->item->item = new_token(strlen(op->name) + 1);
		strcpy(conds->item->item, op->name);

		and_end = conds;
	   }
	   else {
		/* must include the action as well as the conditions */
		temp = conds;

		conds = new_fact_list();
		conds->item = new_token_list();
		conds->item->item = new_token(strlen(op->name) + 1);
		strcpy(conds->item->item, op->name);

		conds->next = temp;

		temp = conds;

		conds = make_adl_fact(AND_CONST);
		conds->next = temp;

		for ( ; temp->next; temp = temp->next) {
		  ;
		}

		temp->next = make_adl_fact(RBRACK_CONST);
		and_end = temp->next;
	   }

	   /* now, add this to end and update end */
	   (*end)->next = conds;
	   *end = and_end;
	 }
    }
    assert(get_adl_token(current) == RBRACK_CONST);
    /* skip to the end of the clause */
    clause = current;
  }
}

/*
 * Build frame axioms to keep things the same unless they are changed
 */
void build_frame_axioms()
{
  build_predicate_frame_axioms();
  build_constraint_frame_axioms();
}

/* Build the constraint that sets var' in relation to var as
   needed. If no influences, that's simply var' = var. If influences, 
   that's var' = var + influences. */
token_list build_constraint_for_frame_axiom(token_list var)
{
  token_list tl;
  token_list constraint;
  token_list influences = 0;

  constraint = new_token_list();
  constraint->item = new_token(strlen(EFFECT_STR) + 1);
  strcpy(constraint->item, EFFECT_STR);

  constraint->next = tl = new_token_list();
  tl->item = new_token(strlen(SET_STR) + 1);
  strcpy(tl->item, SET_STR);

  tl->next = new_token_list();
  tl = tl->next;
  tl->item = new_token(strlen(var->item) + 1);
  strcpy(tl->item, var->item);

  tl->next = new_token_list();
  tl = tl->next;
  tl->item = new_token(strlen("=") + 1);
  strcpy(tl->item, "=");

  tl->next = new_token_list();
  tl = tl->next;
  tl->item = new_token(strlen(var->item) + 1);
  strcpy(tl->item, var->item);

  /* If influences are off, we're done */
  if (!bInfluences)
    return constraint;

  /* Otherwise, need to add on all the matching influence variables: */
  for (operator_list op = operators; op; op = op->next) {
    for (fact_list f = op->hooks; f; f = f->next) {
	 if (is_influence(f) && 
		strcmp(var->item, f->item->next->next->next->next->item) == SAME) {
	   /* get the renumbered token from the left side of the '=' */
	   token changed_token = f->item->next->next->item;
	   token_list temp = influences;
	   influences = new_token_list();
	   influences->item = new_token(strlen(changed_token) + 1);
	   strcpy(influences->item, changed_token);
	   influences->next = temp;
	 }
    }
  }

  /* Eliminate repeats */
  for (token_list t1 = influences; t1; t1 = t1->next) {
    token_list back = t1;
    for (token_list t2 = t1->next; t2; t2 = t2->next) {
	 if (strcmp(t1->item, t2->item) == SAME) {
	   back->next = back->next->next;
	   t2->next = NULL;
	   free_complete_token_list(t2);
	 }
	 back = t2;
    }
  }

  /* Add influences onto the constraint */
  while (influences) {
    tl->next = new_token_list();
    tl = tl->next;
    tl->item = new_token(strlen("+") + 1);
    strcpy(tl->item, "+");
    
    tl->next = influences;
    tl = tl->next;
    influences = influences->next;
    tl->next = NULL;
  }

  return constraint;
}

void build_constraint_frame_axioms()
{
  /* !(?v(i+1) = ?v(i)) -> (action with set v & preconds) | ... ) */
  /* (?v(i+1) = ?v(i) | (action & preconds) | ... ) */

  fact_list var, fl, temp, end;
  operator_list i_op;
  int i;

  /* Step through the domain_vars_list looking for fluents */
  for (var = domain_vars_list; var; var = var->next) {
    /* For each fluent, make frame axioms for each step by specifying
	  a constraint keeping the fluent unchanged unless it is the 
	  object of a set statement (look for set statements, if one
	  exists run find_preconds and construct same sort of stmt as
	  in reg frame axioms */
    if (strcmp(var->item->next->item, FLUENT_STR) == SAME) {
	 /* construct the necessary constraint */
	 token_list dummy;
	 token_list constraint = build_constraint_for_frame_axiom(var->item);

	 /* "test" helps find the actions I really want to interfere with the
	    frame axiom. I don't want to catch influences, so I make my test
	    statement an influence. */
	 token_list test = copy_complete_token_list(constraint, &dummy);
	 my_free(test->next->item);
	 test->next->item = new_token(strlen(INFLUENCE_STR) + 1);
	 strcpy(test->next->item, INFLUENCE_STR);

	 /* set up the OR clause with the frame constraint: */
	 fl = make_adl_fact(OR_CONST);
	 fl->next = new_fact_list();
	 fl->next->item = constraint;
	 end = fl->next;

	 for (i_op = operators; i_op; i_op = i_op->next ) {
	   try_add_action_to_frame(i_op, test, &end);
	 }

	 end->next = make_adl_fact(RBRACK_CONST);

	 fl = rewrite_in_cnf(&fl);

	 for (i = 0; i < gproblem_length; i++) {
	   temp = copy_complete_fact_list(fl, &end);
	   time_label_fact_list(temp, i);

	   end->next = NULL;
	   solver_stmt.add(temp);
	   free_complete_fact_list(temp);
	 }

	 free_complete_fact_list(fl);
	 free_complete_token_list(test);
    }
  }
}

void build_predicate_frame_axioms()
{
  /* I assume that any "or" clause an effect is in is a result of a "when" */
  /* v(i) ^ !v(i+1) -> 
	            ((action that sets (!v) and (!v)'s cond preconds) v ... ) */

  fact_list pred, i_list, pos_frame = NULL, neg_frame = NULL, temp, end;
  fact_list neg_curr, pos_curr;
  token_list dummy_tl;
  operator_list i_op;
  int i;

  for (pred = predicates_list; pred; pred = pred->next) {
    /* instantiate the predicate in all possible ways */
    i_list = instantiate_pred( pred->item->item,
						 pred->item->next,
						 NULL );

    /* step through the list, finding actions that make the predicate 
	  true/false and adding them to the frame axiom */
    while (i_list) {
	 token_list pos_effect, neg_effect;

	 pos_effect = new_token_list();
	 pos_effect->item = new_token(strlen(EFFECT_STR) + 1);
	 strcpy(pos_effect->item, EFFECT_STR);
	 pos_effect->next = i_list->item;

	 neg_effect = new_token_list();
	 neg_effect->item = new_token(strlen(EFFECT_STR) + 1);
	 sprintf(neg_effect->item, "%s%s", "!", EFFECT_STR);
	 neg_effect->next = i_list->item;

	 /* set up the pos and neg frames for this pred */

	 /* pos_frame is (!v(i) ^ v(i+1) -> ...), so it looks like OR(v(i), !v(i+1), ... ) */
	 pos_frame = make_adl_fact(OR_CONST);

	 /* PRECOND v */
	 pos_frame->next = new_fact_list();
	 pos_frame->next->item = new_token_list();
	 pos_frame->next->item->item = new_token(strlen(PRECOND_STR) + 1);
	 strcpy(pos_frame->next->item->item, PRECOND_STR);
	 pos_frame->next->item->next = copy_complete_token_list(i_list->item, &dummy_tl);

	 /* !EFFECT v */
	 pos_frame->next->next = new_fact_list();
	 pos_frame->next->next->item = new_token_list();
	 pos_frame->next->next->item->item = new_token(strlen(EFFECT_STR) + 2);
	 sprintf(pos_frame->next->next->item->item, "%s%s", "!", EFFECT_STR);
	 pos_frame->next->next->item->next = copy_complete_token_list(i_list->item, &dummy_tl);

	 /* neg_frame is (v(i) ^ !v(i+1) -> ...), so it looks like OR(!v(i), v(i+1), ... ) */
	 neg_frame = make_adl_fact(OR_CONST);

	 /* !PRECOND v */
	 neg_frame->next = new_fact_list();
	 neg_frame->next->item = new_token_list();
	 neg_frame->next->item->item = new_token(strlen(PRECOND_STR) + 2);
	 sprintf(neg_frame->next->item->item, "%s%s", "!", PRECOND_STR);
	 neg_frame->next->item->next = copy_complete_token_list(i_list->item, &dummy_tl);

	 /* EFFECT v */
	 neg_frame->next->next = new_fact_list();
	 neg_frame->next->next->item = new_token_list();
	 neg_frame->next->next->item->item = new_token(strlen(EFFECT_STR) + 1);
	 strcpy(neg_frame->next->next->item->item, EFFECT_STR);
	 neg_frame->next->next->item->next = copy_complete_token_list(i_list->item, &dummy_tl);

	 pos_curr = pos_frame->next->next;
	 neg_curr = neg_frame->next->next;

	 /* loop through actions looking for any that cause !v or v */
	 for (i_op = operators; i_op; i_op = i_op->next) {
	   try_add_action_to_frame(i_op, pos_effect, &pos_curr);
	   try_add_action_to_frame(i_op, neg_effect, &neg_curr);
	 }

	 pos_curr->next = make_adl_fact(RBRACK_CONST);
	 neg_curr->next = make_adl_fact(RBRACK_CONST);

	 pos_frame = rewrite_in_cnf( &pos_frame );
	 neg_frame = rewrite_in_cnf( &neg_frame );

	 for (i = 0; i < gproblem_length; i++) {
	   temp = copy_complete_fact_list(pos_frame, &end);
	   time_label_fact_list(temp, i);
	   
	   end->next = NULL;
	   solver_stmt.add(temp);
	   free_complete_fact_list(temp);

	   temp = copy_complete_fact_list(neg_frame, &end);
	   time_label_fact_list(temp, i);
	   
	   end->next = NULL;
	   solver_stmt.add(temp);
	   free_complete_fact_list(temp);
	 }

	 free_complete_fact_list(pos_frame);
	 free_complete_fact_list(neg_frame);

	 i_list = i_list->next;
    }
  }
}

/* set up unit clauses for goal axioms at final time step */
void build_goal()
{
  fact_list fl;

  if (!raw_goal_facts)
    return;

  time_label_fact_list( raw_goal_facts, gproblem_length );

  for (fl = raw_goal_facts; fl->next; fl = fl->next) {
    ;
  }

  solver_stmt.add(raw_goal_facts);
  free_complete_fact_list(raw_goal_facts);
}



